/*
 * Copyright (c) 2019 Ori Bernstein <ori@openbsd.org>
 * Copyright (c) 2021 Stefan Sperling <stsp@openbsd.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#define GOT_PKT_MAX	65536

const struct got_error *got_pkt_readn(ssize_t *off, int fd, void *buf,
    size_t n, int timeout);
const struct got_error *got_pkt_flushpkt(int fd, int chattygot);
const struct got_error *got_pkt_readlen(int *len, const char *str, int chattygot);
const struct got_error *got_pkt_readhdr(int *datalen, int fd, int chattygot,
    int timeout);
const struct got_error *got_pkt_readpkt(int *outlen, int fd, char *buf,
    int buflen, int chattygot, int timeout);
const struct got_error *got_pkt_writepkt(int fd, char *buf, int nbuf,
    int chattygot);
